@@ -29,44 +29,52 @@
 */
 
 
-#include <math.h>
+#include <cmath>
 #include <string>
 #include <sstream>
 #include <list>
-#include <boost/tr1/memory.hpp>
-using namespace std::tr1;
 
 #include <OpenEXR/ImathMatrix.h>
 
-#include "dassert.h"
-#include "typedesc.h"
-#include "varyingref.h"
-#include "ustring.h"
-#include "strutil.h"
-#include "hash.h"
-#include "thread.h"
-#include "fmath.h"
-#include "filter.h"
-#include "imageio.h"
+#include <OpenImageIO/dassert.h>
+#include <OpenImageIO/typedesc.h>
+#include <OpenImageIO/varyingref.h>
+#include <OpenImageIO/ustring.h>
+#include <OpenImageIO/strutil.h>
+#include <OpenImageIO/thread.h>
+#include <OpenImageIO/fmath.h>
+#include <OpenImageIO/filter.h>
+#include <OpenImageIO/imageio.h>
+#include <OpenImageIO/texture.h>
+#include <OpenImageIO/imagecache.h>
 
-#include "texture.h"
-
-#include "imagecache.h"
 #include "imagecache_pvt.h"
 #include "texture_pvt.h"
-#include "../field3d.imageio/field3d_pvt.h"
+#include "../field3d.imageio/field3d_backdoor.h"
 
-OIIO_NAMESPACE_ENTER
-{
-    using namespace pvt;
+OIIO_NAMESPACE_BEGIN
+using namespace pvt;
+using namespace f3dpvt;
 
 namespace {  // anonymous
 
 static EightBitConverter<float> uchar2float;
 static ustring s_field3d ("field3d");
 
+// OIIO_FORCEINLINE float uchar2float (unsigned char val) {
+//     return float(val) * (1.0f/255.0f);
+// }
 
-};  // end anonymous namespace
+OIIO_FORCEINLINE float ushort2float (unsigned short val) {
+    return float(val) * (1.0f/65535.0f);
+}
+
+OIIO_FORCEINLINE float half2float (half val) {
+    return float(val);
+}
+
+
+}  // end anonymous namespace
 
 namespace pvt {   // namespace pvt
 
@@ -78,12 +86,15 @@ TextureSystemImpl::texture3d (ustring filename, TextureOpt &options,
                               const Imath::V3f &dPdx,
                               const Imath::V3f &dPdy,
                               const Imath::V3f &dPdz,
-                              float *result)
+                              int nchannels, float *result,
+                              float *dresultds, float *dresultdt,
+                              float *dresultdr)
 {
     PerThreadInfo *thread_info = m_imagecache->get_perthread_info ();
     TextureFile *texturefile = find_texturefile (filename, thread_info);
     return texture3d ((TextureHandle *)texturefile, (Perthread *)thread_info,
-                      options, P, dPdx, dPdy, dPdz, result);
+                      options, P, dPdx, dPdy, dPdz,
+                      nchannels, result, dresultds, dresultdt);
 }
 
 
@@ -96,8 +107,31 @@ TextureSystemImpl::texture3d (TextureHandle *texture_handle_,
                               const Imath::V3f &dPdx,
                               const Imath::V3f &dPdy,
                               const Imath::V3f &dPdz,
-                              float *result)
+                              int nchannels, float *result,
+                              float *dresultds, float *dresultdt,
+                              float *dresultdr)
 {
+    // Handle >4 channel lookups by recursion.
+    if (nchannels > 4) {
+        int save_firstchannel = options.firstchannel;
+        while (nchannels) {
+            int n = std::min (nchannels, 4);
+            bool ok = texture3d (texture_handle_, thread_info_, options,
+                                 P, dPdx, dPdy, dPdz,
+                                 n, result, dresultds, dresultdt, dresultdr);
+            if (! ok)
+                return false;
+            result += n;
+            if (dresultds) dresultds += n;
+            if (dresultdt) dresultdt += n;
+            if (dresultdr) dresultdr += n;
+            options.firstchannel += n;
+            nchannels -= n;
+        }
+        options.firstchannel = save_firstchannel; // restore what we changed
+        return true;
+    }
+
 #if 0
     // FIXME: currently, no support of actual MIPmapping.  No rush,
     // since the only volume format we currently support, Field3D,
@@ -115,40 +149,46 @@ TextureSystemImpl::texture3d (TextureHandle *texture_handle_,
     texture3d_lookup_prototype lookup = &TextureSystemImpl::texture3d_lookup_nomip;
 #endif
 
-    PerThreadInfo *thread_info = (PerThreadInfo *)thread_info_;
-    TextureFile *texturefile = (TextureFile *)texture_handle_;
+    PerThreadInfo *thread_info = m_imagecache->get_perthread_info((PerThreadInfo *)thread_info_);
+    TextureFile *texturefile = verify_texturefile ((TextureFile *)texture_handle_, thread_info);
     ImageCacheStatistics &stats (thread_info->m_stats);
     ++stats.texture3d_batches;
     ++stats.texture3d_queries;
 
     if (! texturefile  ||  texturefile->broken())
-        return missing_texture (options, result);
+        return missing_texture (options, nchannels, result,
+                                dresultds, dresultdt, dresultdr);
+
+    if (options.subimagename) {
+        // If subimage was specified by name, figure out its index.
+        int s = m_imagecache->subimage_from_name (texturefile, options.subimagename);
+        if (s < 0) {
+            error ("Unknown subimage \"%s\" in texture \"%s\"",
+                   options.subimagename.c_str(), texturefile->filename().c_str());
+            return false;
+        }
+        options.subimage = s;
+        options.subimagename.clear();
+    }
 
     const ImageSpec &spec (texturefile->spec(options.subimage, 0));
 
     // Figure out the wrap functions
     if (options.swrap == TextureOpt::WrapDefault)
-        options.swrap = texturefile->swrap();
-    if (options.swrap == TextureOpt::WrapPeriodic && ispow2(spec.full_width))
-        options.swrap_func = wrap_periodic2;
-    else
-        options.swrap_func = wrap_functions[(int)options.swrap];
+        options.swrap = (TextureOpt::Wrap)texturefile->swrap();
+    if (options.swrap == TextureOpt::WrapPeriodic && ispow2(spec.width))
+        options.swrap = TextureOpt::WrapPeriodicPow2;
     if (options.twrap == TextureOpt::WrapDefault)
-        options.twrap = texturefile->twrap();
-    if (options.twrap == TextureOpt::WrapPeriodic && ispow2(spec.full_height))
-        options.twrap_func = wrap_periodic2;
-    else
-        options.twrap_func = wrap_functions[(int)options.twrap];
+        options.twrap = (TextureOpt::Wrap)texturefile->twrap();
+    if (options.twrap == TextureOpt::WrapPeriodic && ispow2(spec.height))
+        options.twrap = TextureOpt::WrapPeriodicPow2;
     if (options.rwrap == TextureOpt::WrapDefault)
-        options.rwrap = texturefile->rwrap();
-    if (options.rwrap == TextureOpt::WrapPeriodic && ispow2(spec.full_depth))
-        options.rwrap_func = wrap_periodic2;
-    else
-        options.rwrap_func = wrap_functions[(int)options.rwrap];
+        options.rwrap = (TextureOpt::Wrap)texturefile->rwrap();
+    if (options.rwrap == TextureOpt::WrapPeriodic && ispow2(spec.depth))
+        options.rwrap = TextureOpt::WrapPeriodicPow2;
 
     int actualchannels = Imath::clamp (spec.nchannels - options.firstchannel,
-                                       0, options.nchannels);
-    options.actualchannels = actualchannels;
+                                       0, nchannels);
 
     // Do the volume lookup in local space.  There's not actually a way
     // to ask for point transforms via the ImageInput interface, so use
@@ -173,12 +213,13 @@ TextureSystemImpl::texture3d (TextureHandle *texture_handle_,
     // need to transform them into local space as well.
 
     bool ok = (this->*lookup) (*texturefile, thread_info, options,
-                               Plocal, dPdx, dPdy, dPdz, result);
-    if (actualchannels < options.nchannels)
-        fill_channels (spec, options, result);
-    return ok;
-
+                               nchannels, actualchannels,
+                               Plocal, dPdx, dPdy, dPdz,
+                               result, dresultds, dresultdt, dresultdr);
 
+    if (actualchannels < nchannels && options.firstchannel == 0 && m_gray_to_rgb)
+        fill_gray_channels (spec, nchannels, result,
+                            dresultds, dresultdt, dresultdr);
     return ok;
 }
 
@@ -191,14 +232,50 @@ TextureSystemImpl::texture3d (ustring filename, TextureOptions &options,
                               VaryingRef<Imath::V3f> dPdx,
                               VaryingRef<Imath::V3f> dPdy,
                               VaryingRef<Imath::V3f> dPdz,
-                              float *result)
+                              int nchannels, float *result,
+                              float *dresultds, float *dresultdt,
+                              float *dresultdr)
+{
+    Perthread *thread_info = get_perthread_info();
+    TextureHandle *texture_handle = get_texture_handle (filename, thread_info);
+    return texture3d (texture_handle, thread_info, options,
+                      runflags, beginactive, endactive,
+                      P, dPdx, dPdy, dPdz,
+                      nchannels, result, dresultds, dresultdt, dresultdr);
+}
+
+
+
+bool
+TextureSystemImpl::texture3d (TextureHandle *texture_handle,
+                              Perthread *thread_info, TextureOptions &options,
+                              Runflag *runflags, int beginactive, int endactive,
+                              VaryingRef<Imath::V3f> P,
+                              VaryingRef<Imath::V3f> dPdx,
+                              VaryingRef<Imath::V3f> dPdy,
+                              VaryingRef<Imath::V3f> dPdz,
+                              int nchannels, float *result,
+                              float *dresultds, float *dresultdt,
+                              float *dresultdr)
 {
     bool ok = true;
+    result += beginactive*nchannels;
+    if (dresultds) {
+        dresultds += beginactive*nchannels;
+        dresultdt += beginactive*nchannels;
+    }
     for (int i = beginactive;  i < endactive;  ++i) {
         if (runflags[i]) {
             TextureOpt opt (options, i);
-            ok &= texture3d (filename, opt, P[i], dPdx[i], dPdy[i], dPdz[i],
-                             result + i*options.nchannels);
+            ok &= texture3d (texture_handle, thread_info, opt,
+                             P[i], dPdx[i], dPdy[i], dPdz[i],
+                             4, result, dresultds, dresultdt, dresultdr);
+        }
+        result += nchannels;
+        if (dresultds) {
+            dresultds += nchannels;
+            dresultdt += nchannels;
+            dresultdr += nchannels;
         }
     }
     return ok;
@@ -210,19 +287,24 @@ bool
 TextureSystemImpl::texture3d_lookup_nomip (TextureFile &texturefile,
                             PerThreadInfo *thread_info, 
                             TextureOpt &options,
-                            const Imath::V3f &_P, const Imath::V3f &_dPdx,
-                            const Imath::V3f &_dPdy, const Imath::V3f &_dPdz,
-                            float *result)
+                            int nchannels_result, int actualchannels,
+                            const Imath::V3f &P, const Imath::V3f &dPdx,
+                            const Imath::V3f &dPdy, const Imath::V3f &dPdz,
+                            float *result,
+                            float *dresultds, float *dresultdt,
+                            float *dresultdr)
 {
     // Initialize results to 0.  We'll add from here on as we sample.
-    float* dresultds = options.dresultds;
-    float* dresultdt = options.dresultdt;
-    float* dresultdr = options.dresultdr;
-    for (int c = 0;  c < options.actualchannels;  ++c) {
+    for (int c = 0;  c < nchannels_result;  ++c)
         result[c] = 0;
-        if (dresultds) dresultds[c] = 0;
-        if (dresultdt) dresultdt[c] = 0;
-        if (dresultdr) dresultdr[c] = 0;
+    if (dresultds) {
+        DASSERT (dresultdt && dresultdr);
+        for (int c = 0;  c < nchannels_result;  ++c)
+            dresultds[c] = 0;
+        for (int c = 0;  c < nchannels_result;  ++c)
+            dresultdt[c] = 0;
+        for (int c = 0;  c < nchannels_result;  ++c)
+            dresultdr[c] = 0;
     }
     // If the user only provided us with one pointer, clear all to simplify
     // the rest of the code, but only after we zero out the data for them so
@@ -238,7 +320,8 @@ TextureSystemImpl::texture3d_lookup_nomip (TextureFile &texturefile,
         &TextureSystemImpl::accum3d_sample_bilinear,
     };
     accum3d_prototype accumer = accum_functions[(int)options.interpmode];
-    bool ok = (this->*accumer) (_P, 0, texturefile, thread_info, options,
+    bool ok = (this->*accumer) (P, 0, texturefile, thread_info, options,
+                                nchannels_result, actualchannels,
                                 1.0f, result, dresultds, dresultdt, dresultdr);
 
     // Update stats
@@ -261,11 +344,13 @@ TextureSystemImpl::accum3d_sample_closest (const Imath::V3f &P, int miplevel,
                                  TextureFile &texturefile,
                                  PerThreadInfo *thread_info,
                                  TextureOpt &options,
+                                 int nchannels_result, int actualchannels,
                                  float weight, float *accum, float *daccumds,
                                  float *daccumdt, float *daccumdr)
 {
     const ImageSpec &spec (texturefile.spec (options.subimage, miplevel));
     const ImageCacheFile::LevelInfo &levelinfo (texturefile.levelinfo(options.subimage,miplevel));
+    TypeDesc::BASETYPE pixeltype = texturefile.pixeltype(options.subimage);
     // As passed in, (s,t) map the texture to (0,1).  Remap to texel coords.
     float s = P[0] * spec.full_width  + spec.full_x;
     float t = P[1] * spec.full_height + spec.full_y;
@@ -275,13 +360,13 @@ TextureSystemImpl::accum3d_sample_closest (const Imath::V3f &P, int miplevel,
     (void) floorfrac (t, &ttex);
     (void) floorfrac (r, &rtex);
 
-    // Wrap
-    DASSERT (options.swrap_func != NULL && options.twrap_func != NULL &&
-             options.rwrap_func != NULL);
+    wrap_impl swrap_func = wrap_functions[(int)options.swrap];
+    wrap_impl twrap_func = wrap_functions[(int)options.twrap];
+    wrap_impl rwrap_func = wrap_functions[(int)options.rwrap];
     bool svalid, tvalid, rvalid;  // Valid texels?  false means black border
-    svalid = options.swrap_func (stex, spec.x, spec.width);
-    tvalid = options.twrap_func (ttex, spec.y, spec.height);
-    rvalid = options.rwrap_func (rtex, spec.z, spec.depth);
+    svalid = swrap_func (stex, spec.x, spec.width);
+    tvalid = twrap_func (ttex, spec.y, spec.height);
+    rvalid = rwrap_func (rtex, spec.z, spec.depth);
     if (! levelinfo.full_pixel_range) {
         svalid &= (stex >= spec.x && stex < (spec.x+spec.width)); // data window
         tvalid &= (ttex >= spec.y && ttex < (spec.y+spec.height));
@@ -292,35 +377,54 @@ TextureSystemImpl::accum3d_sample_closest (const Imath::V3f &P, int miplevel,
         return true;
     }
 
-    int tilewidthmask  = spec.tile_width  - 1;  // e.g. 63
-    int tileheightmask = spec.tile_height - 1;
-    int tiledepthmask = spec.tile_depth - 1;
-    int tile_s = (stex - spec.x) & tilewidthmask;
-    int tile_t = (ttex - spec.y) & tileheightmask;
-    int tile_r = (rtex - spec.z) & tiledepthmask;
+    int tile_chbegin = 0, tile_chend = spec.nchannels;
+    if (spec.nchannels > m_max_tile_channels) {
+        // For files with many channels, narrow the range we cache
+        tile_chbegin = options.firstchannel;
+        tile_chend = options.firstchannel+actualchannels;
+    }
+    int tile_s = (stex - spec.x) % spec.tile_width;
+    int tile_t = (ttex - spec.y) % spec.tile_height;
+    int tile_r = (rtex - spec.z) % spec.tile_depth;
     TileID id (texturefile, options.subimage, miplevel,
-               stex - tile_s, ttex - tile_t, rtex - tile_r);
+               stex - tile_s, ttex - tile_t, rtex - tile_r,
+               tile_chbegin, tile_chend);
     bool ok = find_tile (id, thread_info);
     if (! ok)
         error ("%s", m_imagecache->geterror().c_str());
     TileRef &tile (thread_info->tile);
     if (! tile  ||  ! ok)
         return false;
-    size_t channelsize = texturefile.channelsize();
     int tilepel = (tile_r * spec.tile_height + tile_t) * spec.tile_width + tile_s;
-    int offset = spec.nchannels * tilepel + options.firstchannel;
+    int startchan_in_tile = options.firstchannel - id.chbegin();
+    int offset = spec.nchannels * tilepel + startchan_in_tile;
     DASSERT ((size_t)offset < spec.nchannels*spec.tile_pixels());
-    if (channelsize == 1) {
-        // special case for 8-bit tiles
+    if (pixeltype == TypeDesc::UINT8) {
         const unsigned char *texel = tile->bytedata() + offset;
-        for (int c = 0;  c < options.actualchannels;  ++c)
+        for (int c = 0;  c < actualchannels;  ++c)
             accum[c] += weight * uchar2float(texel[c]);
+    } else if (pixeltype == TypeDesc::UINT16) {
+        const unsigned short *texel = tile->ushortdata() + offset;
+        for (int c = 0;  c < actualchannels;  ++c)
+            accum[c] += weight * ushort2float(texel[c]);
+    } else if (pixeltype == TypeDesc::HALF) {
+        const half *texel = tile->halfdata() + offset;
+        for (int c = 0;  c < actualchannels;  ++c)
+            accum[c] += weight * half2float(texel[c]);
     } else {
-        // General case for float tiles
-        const float *texel = tile->data() + offset;
-        for (int c = 0;  c < options.actualchannels;  ++c)
+        DASSERT (pixeltype == TypeDesc::FLOAT);
+        const float *texel = tile->floatdata() + offset;
+        for (int c = 0;  c < actualchannels;  ++c)
             accum[c] += weight * texel[c];
     }
+
+    // Add appropriate amount of "fill" color to extra channels in
+    // non-"black"-wrapped regions.
+    if (nchannels_result > actualchannels && options.fill) {
+        float f = weight * options.fill;
+        for (int c = actualchannels;  c < nchannels_result;  ++c)
+            accum[c] += f;
+    }
     return true;
 }
 
@@ -331,11 +435,13 @@ TextureSystemImpl::accum3d_sample_bilinear (const Imath::V3f &P, int miplevel,
                                  TextureFile &texturefile,
                                  PerThreadInfo *thread_info,
                                  TextureOpt &options,
+                                 int nchannels_result, int actualchannels,
                                  float weight, float *accum, float *daccumds,
                                  float *daccumdt, float *daccumdr)
 {
     const ImageSpec &spec (texturefile.spec (options.subimage, miplevel));
     const ImageCacheFile::LevelInfo &levelinfo (texturefile.levelinfo(options.subimage,miplevel));
+    TypeDesc::BASETYPE pixeltype = texturefile.pixeltype(options.subimage);
     // As passed in, (s,t) map the texture to (0,1).  Remap to texel coords
     // and subtract 0.5 because samples are at texel centers.
     float s = P[0] * spec.full_width  + spec.full_x - 0.5f;
@@ -352,8 +458,10 @@ TextureSystemImpl::accum3d_sample_bilinear (const Imath::V3f &P, int miplevel,
     // and to the right).
 
     // Wrap
-    DASSERT (options.swrap_func != NULL && options.twrap_func != NULL &&
-             options.rwrap_func != NULL);
+    wrap_impl swrap_func = wrap_functions[(int)options.swrap];
+    wrap_impl twrap_func = wrap_functions[(int)options.twrap];
+    wrap_impl rwrap_func = wrap_functions[(int)options.rwrap];
+
     int stex[2], ttex[2], rtex[2];       // Texel coords
     stex[0] = sint;  stex[1] = sint+1;
     ttex[0] = tint;  ttex[1] = tint+1;
@@ -361,20 +469,19 @@ TextureSystemImpl::accum3d_sample_bilinear (const Imath::V3f &P, int miplevel,
 //    bool svalid[2], tvalid[2], rvalid[2];  // Valid texels?  false means black border
     union { bool bvalid[6]; unsigned long long ivalid; } valid_storage;
     valid_storage.ivalid = 0;
-    DASSERT (sizeof(valid_storage) >= 6*sizeof(bool));
+    DASSERT (sizeof(valid_storage) == 8);
     const unsigned long long none_valid = 0;
-    const unsigned long long all_valid = 0x010101010101LL;
-    DASSERT (__LITTLE_ENDIAN__ && "this trick won't work with big endian");
+    const unsigned long long all_valid = littleendian() ? 0x010101010101LL : 0x01010101010100LL;
     bool *svalid = valid_storage.bvalid;
     bool *tvalid = valid_storage.bvalid + 2;
     bool *rvalid = valid_storage.bvalid + 4;
 
-    svalid[0] = options.swrap_func (stex[0], spec.x, spec.width);
-    svalid[1] = options.swrap_func (stex[1], spec.x, spec.width);
-    tvalid[0] = options.twrap_func (ttex[0], spec.y, spec.height);
-    tvalid[1] = options.twrap_func (ttex[1], spec.y, spec.height);
-    rvalid[0] = options.rwrap_func (rtex[0], spec.z, spec.depth);
-    rvalid[1] = options.rwrap_func (rtex[1], spec.z, spec.depth);
+    svalid[0] = swrap_func (stex[0], spec.x, spec.width);
+    svalid[1] = swrap_func (stex[1], spec.x, spec.width);
+    tvalid[0] = twrap_func (ttex[0], spec.y, spec.height);
+    tvalid[1] = twrap_func (ttex[1], spec.y, spec.height);
+    rvalid[0] = rwrap_func (rtex[0], spec.z, spec.depth);
+    rvalid[1] = rwrap_func (rtex[1], spec.z, spec.depth);
     // Account for crop windows
     if (! levelinfo.full_pixel_range) {
         svalid[0] &= (stex[0] >= spec.x && stex[0] < spec.x+spec.width);
@@ -394,20 +501,28 @@ TextureSystemImpl::accum3d_sample_bilinear (const Imath::V3f &P, int miplevel,
     const unsigned char *texel[2][2][2];
     TileRef savetile[2][2][2];
     static float black[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
-    int tile_s = (stex[0] - spec.x) & tilewidthmask;
-    int tile_t = (ttex[0] - spec.y) & tileheightmask;
-    int tile_r = (rtex[0] - spec.z) & tiledepthmask;
+    int tile_s = (stex[0] - spec.x) % spec.tile_width;
+    int tile_t = (ttex[0] - spec.y) % spec.tile_height;
+    int tile_r = (rtex[0] - spec.z) % spec.tile_depth;
     bool s_onetile = (tile_s != tilewidthmask) & (stex[0]+1 == stex[1]);
     bool t_onetile = (tile_t != tileheightmask) & (ttex[0]+1 == ttex[1]);
     bool r_onetile = (tile_r != tiledepthmask) & (rtex[0]+1 == rtex[1]);
     bool onetile = (s_onetile & t_onetile & r_onetile);
-    size_t channelsize = texturefile.channelsize();
-    size_t pixelsize = texturefile.pixelsize();
+    size_t channelsize = texturefile.channelsize(options.subimage);
+    size_t pixelsize = texturefile.pixelsize(options.subimage);
+    int tile_chbegin = 0, tile_chend = spec.nchannels;
+    if (spec.nchannels > m_max_tile_channels) {
+        // For files with many channels, narrow the range we cache
+        tile_chbegin = options.firstchannel;
+        tile_chend = options.firstchannel+actualchannels;
+    }
+    TileID id (texturefile, options.subimage, miplevel, 0, 0, 0,
+               tile_chbegin, tile_chend);
+    int startchan_in_tile = options.firstchannel - id.chbegin();
     if (onetile &&
         valid_storage.ivalid == all_valid) {
         // Shortcut if all the texels we need are on the same tile
-        TileID id (texturefile, options.subimage, miplevel,
-                   stex[0] - tile_s, ttex[0] - tile_t, rtex[0] - tile_r);
+        id.xyz (stex[0] - tile_s, ttex[0] - tile_t, rtex[0] - tile_r);
         bool ok = find_tile (id, thread_info);
         if (! ok)
             error ("%s", m_imagecache->geterror().c_str());
@@ -415,7 +530,7 @@ TextureSystemImpl::accum3d_sample_bilinear (const Imath::V3f &P, int miplevel,
         if (! tile->valid())
             return false;
         size_t tilepel = (tile_r * spec.tile_height + tile_t) * spec.tile_width + tile_s;
-        size_t offset = (spec.nchannels * tilepel + options.firstchannel) * channelsize;
+        size_t offset = (spec.nchannels * tilepel + startchan_in_tile) * channelsize;
         DASSERT ((size_t)offset < spec.tile_width*spec.tile_height*spec.tile_depth*pixelsize);
 
         const unsigned char *b = tile->bytedata() + offset;
@@ -436,12 +551,11 @@ TextureSystemImpl::accum3d_sample_bilinear (const Imath::V3f &P, int miplevel,
                         texel[k][j][i] = (unsigned char *)black;
                         continue;
                     }
-                    tile_s = (stex[i] - spec.x) & tilewidthmask;
-                    tile_t = (ttex[j] - spec.y) & tileheightmask;
-                    tile_r = (rtex[k] - spec.z) & tiledepthmask;
-                    TileID id (texturefile, options.subimage, miplevel,
-                               stex[i] - tile_s, ttex[j] - tile_t,
-                               rtex[k] - tile_r);
+                    tile_s = (stex[i] - spec.x) % spec.tile_width;
+                    tile_t = (ttex[j] - spec.y) % spec.tile_height;
+                    tile_r = (rtex[k] - spec.z) % spec.tile_depth;
+                    id.xyz (stex[i] - tile_s, ttex[j] - tile_t,
+                            rtex[k] - tile_r);
                     bool ok = find_tile (id, thread_info);
                     if (! ok)
                         error ("%s", m_imagecache->geterror().c_str());
@@ -450,8 +564,8 @@ TextureSystemImpl::accum3d_sample_bilinear (const Imath::V3f &P, int miplevel,
                         return false;
                     savetile[k][j][i] = tile;
                     size_t tilepel = (tile_r * spec.tile_height + tile_t) * spec.tile_width + tile_s;
-                    size_t offset = (spec.nchannels * tilepel + options.firstchannel) * channelsize;
-#if DEBUG
+                    size_t offset = (spec.nchannels * tilepel + startchan_in_tile) * channelsize;
+#ifndef NDEBUG
                     if ((size_t)offset >= spec.tile_width*spec.tile_height*spec.tile_depth*pixelsize)
                         std::cerr << "offset=" << offset << ", whd " << spec.tile_width << ' ' << spec.tile_height << ' ' << spec.tile_depth << " pixsize " << pixelsize << "\n";
 #endif
@@ -464,10 +578,8 @@ TextureSystemImpl::accum3d_sample_bilinear (const Imath::V3f &P, int miplevel,
     }
     // FIXME -- optimize the above loop by unrolling
 
-    if (channelsize == 1) {
-        // special case for 8-bit tiles
-        int c;
-        for (c = 0;  c < options.actualchannels;  ++c)
+    if (pixeltype == TypeDesc::UINT8) {
+        for (int c = 0;  c < actualchannels;  ++c)
             accum[c] += weight * trilerp (uchar2float(texel[0][0][0][c]), uchar2float(texel[0][0][1][c]),
                                           uchar2float(texel[0][1][0][c]), uchar2float(texel[0][1][1][c]),
                                           uchar2float(texel[1][0][0][c]), uchar2float(texel[1][0][1][c]),
@@ -477,11 +589,10 @@ TextureSystemImpl::accum3d_sample_bilinear (const Imath::V3f &P, int miplevel,
             float scalex = weight * spec.full_width;
             float scaley = weight * spec.full_height;
             float scalez = weight * spec.full_depth;
-            for (c = 0;  c < options.actualchannels;  ++c) {
+            for (int c = 0;  c < actualchannels;  ++c) {
                 daccumds[c] += scalex * bilerp(
                     uchar2float(texel[0][0][1][c]) - uchar2float(texel[0][0][0][c]),
                     uchar2float(texel[0][1][1][c]) - uchar2float(texel[0][1][0][c]),
-
                     uchar2float(texel[1][0][1][c]) - uchar2float(texel[1][0][0][c]),
                     uchar2float(texel[1][1][1][c]) - uchar2float(texel[1][1][0][c]),
                     tfrac, rfrac
@@ -502,18 +613,88 @@ TextureSystemImpl::accum3d_sample_bilinear (const Imath::V3f &P, int miplevel,
                 );
             }
         }
+    } else if (pixeltype == TypeDesc::UINT16) {
+        for (int c = 0;  c < actualchannels;  ++c)
+            accum[c] += weight * trilerp (ushort2float(((const uint16_t *)texel[0][0][0])[c]), ushort2float(((const uint16_t *)texel[0][0][1])[c]),
+                                          ushort2float(((const uint16_t *)texel[0][1][0])[c]), ushort2float(((const uint16_t *)texel[0][1][1])[c]),
+                                          ushort2float(((const uint16_t *)texel[1][0][0])[c]), ushort2float(((const uint16_t *)texel[1][0][1])[c]),
+                                          ushort2float(((const uint16_t *)texel[1][1][0])[c]), ushort2float(((const uint16_t *)texel[1][1][1])[c]),
+                                          sfrac, tfrac, rfrac);
+        if (daccumds) {
+            float scalex = weight * spec.full_width;
+            float scaley = weight * spec.full_height;
+            float scalez = weight * spec.full_depth;
+            for (int c = 0;  c < actualchannels;  ++c) {
+                daccumds[c] += scalex * bilerp(
+                    ushort2float(((const uint16_t *)texel[0][0][1])[c]) - ushort2float(((const uint16_t *)texel[0][0][0])[c]),
+                    ushort2float(((const uint16_t *)texel[0][1][1])[c]) - ushort2float(((const uint16_t *)texel[0][1][0])[c]),
+                    ushort2float(((const uint16_t *)texel[1][0][1])[c]) - ushort2float(((const uint16_t *)texel[1][0][0])[c]),
+                    ushort2float(((const uint16_t *)texel[1][1][1])[c]) - ushort2float(((const uint16_t *)texel[1][1][0])[c]),
+                    tfrac, rfrac
+                );
+                daccumdt[c] += scaley * bilerp(
+                    ushort2float(((const uint16_t *)texel[0][1][0])[c]) - ushort2float(((const uint16_t *)texel[0][0][0])[c]),
+                    ushort2float(((const uint16_t *)texel[0][1][1])[c]) - ushort2float(((const uint16_t *)texel[0][0][1])[c]),
+                    ushort2float(((const uint16_t *)texel[1][1][0])[c]) - ushort2float(((const uint16_t *)texel[1][0][0])[c]),
+                    ushort2float(((const uint16_t *)texel[1][1][1])[c]) - ushort2float(((const uint16_t *)texel[1][0][1])[c]),
+                    sfrac, rfrac
+                );
+                daccumdr[c] += scalez * bilerp(
+                    ushort2float(((const uint16_t *)texel[0][1][0])[c]) - ushort2float(((const uint16_t *)texel[1][1][0])[c]),
+                    ushort2float(((const uint16_t *)texel[0][1][1])[c]) - ushort2float(((const uint16_t *)texel[1][1][1])[c]),
+                    ushort2float(((const uint16_t *)texel[0][0][1])[c]) - ushort2float(((const uint16_t *)texel[1][0][0])[c]),
+                    ushort2float(((const uint16_t *)texel[0][1][1])[c]) - ushort2float(((const uint16_t *)texel[1][1][1])[c]),
+                    sfrac, tfrac
+                );
+            }
+        }
+    } else if (pixeltype == TypeDesc::HALF) {
+        for (int c = 0;  c < actualchannels;  ++c)
+            accum[c] += weight * trilerp (half2float(((const half *)texel[0][0][0])[c]), half2float(((const half *)texel[0][0][1])[c]),
+                                          half2float(((const half *)texel[0][1][0])[c]), half2float(((const half *)texel[0][1][1])[c]),
+                                          half2float(((const half *)texel[1][0][0])[c]), half2float(((const half *)texel[1][0][1])[c]),
+                                          half2float(((const half *)texel[1][1][0])[c]), half2float(((const half *)texel[1][1][1])[c]),
+                                          sfrac, tfrac, rfrac);
+        if (daccumds) {
+            float scalex = weight * spec.full_width;
+            float scaley = weight * spec.full_height;
+            float scalez = weight * spec.full_depth;
+            for (int c = 0;  c < actualchannels;  ++c) {
+                daccumds[c] += scalex * bilerp(
+                    half2float(((const half *)texel[0][0][1])[c]) - half2float(((const half *)texel[0][0][0])[c]),
+                    half2float(((const half *)texel[0][1][1])[c]) - half2float(((const half *)texel[0][1][0])[c]),
+                    half2float(((const half *)texel[1][0][1])[c]) - half2float(((const half *)texel[1][0][0])[c]),
+                    half2float(((const half *)texel[1][1][1])[c]) - half2float(((const half *)texel[1][1][0])[c]),
+                    tfrac, rfrac
+                );
+                daccumdt[c] += scaley * bilerp(
+                    half2float(((const half *)texel[0][1][0])[c]) - half2float(((const half *)texel[0][0][0])[c]),
+                    half2float(((const half *)texel[0][1][1])[c]) - half2float(((const half *)texel[0][0][1])[c]),
+                    half2float(((const half *)texel[1][1][0])[c]) - half2float(((const half *)texel[1][0][0])[c]),
+                    half2float(((const half *)texel[1][1][1])[c]) - half2float(((const half *)texel[1][0][1])[c]),
+                    sfrac, rfrac
+                );
+                daccumdr[c] += scalez * bilerp(
+                    half2float(((const half *)texel[0][1][0])[c]) - half2float(((const half *)texel[1][1][0])[c]),
+                    half2float(((const half *)texel[0][1][1])[c]) - half2float(((const half *)texel[1][1][1])[c]),
+                    half2float(((const half *)texel[0][0][1])[c]) - half2float(((const half *)texel[1][0][0])[c]),
+                    half2float(((const half *)texel[0][1][1])[c]) - half2float(((const half *)texel[1][1][1])[c]),
+                    sfrac, tfrac
+                );
+            }
+        }
     } else {
         // General case for float tiles
         trilerp_mad ((const float *)texel[0][0][0], (const float *)texel[0][0][1],
                      (const float *)texel[0][1][0], (const float *)texel[0][1][1],
                      (const float *)texel[1][0][0], (const float *)texel[1][0][1],
                      (const float *)texel[1][1][0], (const float *)texel[1][1][1],
-                     sfrac, tfrac, rfrac, weight, options.actualchannels, accum);
+                     sfrac, tfrac, rfrac, weight, actualchannels, accum);
         if (daccumds) {
             float scalex = weight * spec.full_width;
             float scaley = weight * spec.full_height;
             float scalez = weight * spec.full_depth;
-            for (int c = 0;  c < options.actualchannels;  ++c) {
+            for (int c = 0;  c < actualchannels;  ++c) {
                 daccumds[c] += scalex * bilerp(
                     ((const float *) texel[0][0][1])[c] - ((const float *) texel[0][0][0])[c],
                     ((const float *) texel[0][1][1])[c] - ((const float *) texel[0][1][0])[c],
@@ -539,12 +720,104 @@ TextureSystemImpl::accum3d_sample_bilinear (const Imath::V3f &P, int miplevel,
         }
     }
 
+    // Add appropriate amount of "fill" color to extra channels in
+    // non-"black"-wrapped regions.
+    if (nchannels_result > actualchannels && options.fill) {
+        float f = trilerp (1.0f*(rvalid[0]*tvalid[0]*svalid[0]), 1.0f*(rvalid[0]*tvalid[0]*svalid[1]),
+                           1.0f*(rvalid[0]*tvalid[1]*svalid[0]), 1.0f*(rvalid[0]*tvalid[1]*svalid[1]),
+                           1.0f*(rvalid[1]*tvalid[0]*svalid[0]), 1.0f*(rvalid[1]*tvalid[0]*svalid[1]),
+                           1.0f*(rvalid[1]*tvalid[1]*svalid[0]), 1.0f*(rvalid[1]*tvalid[1]*svalid[1]),
+                           sfrac, tfrac, rfrac);
+        f *= weight * options.fill;
+        for (int c = actualchannels;  c < nchannels_result;  ++c)
+            accum[c] += f;
+    }
     return true;
 }
 
 
 
-};  // end namespace pvt
+bool
+TextureSystemImpl::texture3d (TextureHandle *texture_handle,
+                              Perthread *thread_info,
+                              TextureOptBatch &options, Tex::RunMask mask,
+                              const float *P, const float *dPdx,
+                              const float *dPdy, const float *dPdz,
+                              int nchannels, float *result,
+                              float *dresultds, float *dresultdt,
+                              float *dresultdr)
+{
+    // (FIXME) CHEAT! Texture points individually
+    TextureOpt opt;
+    opt.firstchannel = options.firstchannel;
+    opt.subimage = options.subimage;
+    opt.subimagename = options.subimagename;
+    opt.swrap = (TextureOpt::Wrap) options.swrap;
+    opt.twrap = (TextureOpt::Wrap) options.twrap;
+    opt.mipmode = (TextureOpt::MipMode) options.mipmode;
+    opt.interpmode = (TextureOpt::InterpMode) options.interpmode;
+    opt.anisotropic = options.anisotropic;
+    opt.conservative_filter = options.conservative_filter;
+    opt.fill = options.fill;
+    opt.missingcolor = options.missingcolor;
+    opt.rwrap = (TextureOpt::Wrap) options.rwrap;
+
+    bool ok = true;
+    Tex::RunMask bit = 1;
+    for (int i = 0; i < Tex::BatchWidth;  ++i, bit <<= 1) {
+        float r[4], drds[4], drdt[4];  // temp result
+        if (mask & bit) {
+            opt.sblur = options.sblur[i];
+            opt.tblur = options.tblur[i];
+            opt.rblur = options.rblur[i];
+            opt.swidth = options.swidth[i];
+            opt.twidth = options.twidth[i];
+            opt.rwidth = options.rwidth[i];
+            Imath::V3f P_ (P[i], P[i+Tex::BatchWidth], P[i+2*Tex::BatchWidth]);
+            Imath::V3f dPdx_ (dPdx[i], dPdx[i+Tex::BatchWidth], dPdx[i+2*Tex::BatchWidth]);
+            Imath::V3f dPdy_ (dPdy[i], dPdy[i+Tex::BatchWidth], dPdy[i+2*Tex::BatchWidth]);
+            Imath::V3f dPdz_ (dPdz[i], dPdz[i+Tex::BatchWidth], dPdz[i+2*Tex::BatchWidth]);
+            if (dresultds) {
+                ok &= texture3d (texture_handle, thread_info, opt,
+                                 P_, dPdx_, dPdy_, dPdz_,
+                                 nchannels, r, drds, drdt);
+                for (int c = 0; c < nchannels; ++c) {
+                    result[c*Tex::BatchWidth+i] = r[c];
+                    dresultds[c*Tex::BatchWidth+i] = drds[c];
+                    dresultdt[c*Tex::BatchWidth+i] = drdt[c];
+                }
+            } else {
+                ok &= texture3d (texture_handle, thread_info, opt,
+                                 P_, dPdx_, dPdy_, dPdz_, nchannels, r);
+                for (int c = 0; c < nchannels; ++c) {
+                    result[c*Tex::BatchWidth+i] = r[c];
+                }
+            }
+        }
+    }
+    return ok;
+}
+
+
+
+bool
+TextureSystemImpl::texture3d (ustring filename,
+                              TextureOptBatch &options, Tex::RunMask mask,
+                              const float *P, const float *dPdx,
+                              const float *dPdy, const float *dPdz,
+                              int nchannels, float *result,
+                              float *dresultds, float *dresultdt,
+                              float *dresultdr)
+{
+    Perthread *thread_info = get_perthread_info();
+    TextureHandle *texture_handle = get_texture_handle (filename, thread_info);
+    return texture3d (texture_handle, thread_info, options, mask,
+                      P, dPdx, dPdy, dPdz, nchannels,
+                      result, dresultds, dresultdt, dresultdr);
 
 }
-OIIO_NAMESPACE_EXIT
+
+
+}  // end namespace pvt
+
+OIIO_NAMESPACE_END
