@@ -32,17 +32,20 @@
 #define OPENIMAGEIO_PNG_PVT_H
 
 #include <png.h>
-
+#include <zlib.h>
 #include <OpenEXR/ImathColor.h>
 
-#include "dassert.h"
-#include "typedesc.h"
-#include "imageio.h"
-#include "strutil.h"
-#include "fmath.h"
-#include "sysutil.h"
+#include <OpenImageIO/dassert.h>
+#include <OpenImageIO/typedesc.h>
+#include <OpenImageIO/imageio.h>
+#include <OpenImageIO/strutil.h>
+#include <OpenImageIO/filesystem.h>
+#include <OpenImageIO/fmath.h>
+#include <OpenImageIO/sysutil.h>
 
 
+#define OIIO_LIBPNG_VERSION (PNG_LIBPNG_VER_MAJOR*10000 + PNG_LIBPNG_VER_MINOR*100 + PNG_LIBPNG_VER_RELEASE)
+
 
 /*
 This code has been extracted from the PNG plugin so as to provide access to PNG
@@ -59,6 +62,9 @@ http://lists.openimageio.org/pipermail/oiio-dev-openimageio.org/2009-April/00065
 
 OIIO_PLUGIN_NAMESPACE_BEGIN
 
+#define ICC_PROFILE_ATTR "ICCProfile"
+
+
 namespace PNG_pvt {
 
 /// Initializes a PNG read struct.
@@ -123,8 +129,8 @@ get_background (png_structp& sp, png_infop& ip, ImageSpec& spec,
 ///
 inline void
 read_info (png_structp& sp, png_infop& ip, int& bit_depth, int& color_type,
-           int& interlace_type,
-           Imath::Color3f& bg, ImageSpec& spec)
+           int& interlace_type, Imath::Color3f& bg, ImageSpec& spec,
+           bool keep_unassociated_alpha)
 {
     png_read_info (sp, ip);
 
@@ -148,19 +154,52 @@ read_info (png_structp& sp, png_infop& ip, int& bit_depth, int& color_type,
                        bit_depth == 16 ? TypeDesc::UINT16 : TypeDesc::UINT8);
 
     spec.default_channel_names ();
-
-    double gamma;
-    if (png_get_gAMA (sp, ip, &gamma)) {
-        spec.attribute ("oiio:Gamma", (float) gamma);
-        spec.attribute ("oiio:ColorSpace", (gamma == 1) ? "Linear" : "GammaCorrected");
+    if (spec.nchannels == 2) {
+        // Special case: PNG spec says 2-channel image is Gray & Alpha
+        spec.channelnames[0] = "Y";
+        spec.channelnames[1] = "A";
+        spec.alpha_channel = 1;
     }
+
     int srgb_intent;
     if (png_get_sRGB (sp, ip, &srgb_intent)) {
         spec.attribute ("oiio:ColorSpace", "sRGB");
+    } else {
+        double gamma;
+        if (png_get_gAMA (sp, ip, &gamma)) {
+            // Round gamma to the nearest hundredth to prevent stupid
+            // precision choices and make it easier for apps to make
+            // decisions based on known gamma values. For example, you want
+            // 2.2, not 2.19998.
+            float g = float (1.0 / gamma);
+            g = roundf (100.0 * g) / 100.0f;
+            spec.attribute ("oiio:Gamma", g);
+            if (g == 1.0f)
+                spec.attribute ("oiio:ColorSpace", "linear");
+            else
+                spec.attribute ("oiio:ColorSpace",
+                                Strutil::format("GammaCorrected%.2g", g));
+        }
     }
+
+    if (png_get_valid(sp, ip, PNG_INFO_iCCP)) {
+        png_charp profile_name = NULL;
+#if OIIO_LIBPNG_VERSION > 10500   /* PNG function signatures changed */
+        png_bytep profile_data = NULL;
+#else
+        png_charp profile_data = NULL;
+#endif
+        png_uint_32 profile_length = 0;
+        int compression_type;
+        png_get_iCCP (sp, ip, &profile_name, &compression_type,
+                      &profile_data, &profile_length);
+        if (profile_length && profile_data)
+            spec.attribute (ICC_PROFILE_ATTR, TypeDesc(TypeDesc::UINT8, profile_length), profile_data);
+    }
+
     png_timep mod_time;
     if (png_get_tIME (sp, ip, &mod_time)) {
-        std::string date = Strutil::format ("%4d:%02d:%02d %2d:%02d:%02d",
+        std::string date = Strutil::format ("%4d:%02d:%02d %02d:%02d:%02d",
                            mod_time->year, mod_time->month, mod_time->day,
                            mod_time->hour, mod_time->minute, mod_time->second);
         spec.attribute ("DateTime", date); 
@@ -177,6 +216,8 @@ read_info (png_structp& sp, png_infop& ip, int& bit_depth, int& color_type,
                 spec.attribute ("Artist", text_ptr[i].text);
             else if (Strutil::iequals (text_ptr[i].key, "Title"))
                 spec.attribute ("DocumentName", text_ptr[i].text);
+            else if (Strutil::iequals (text_ptr[i].key, "XML:com.adobe.xmp"))
+                decode_xmp (text_ptr[i].text, spec);
             else
                 spec.attribute (text_ptr[i].key, text_ptr[i].text);
         }
@@ -210,8 +251,10 @@ read_info (png_structp& sp, png_infop& ip, int& bit_depth, int& color_type,
 
     interlace_type = png_get_interlace_type (sp, ip);
 
-    // PNG files are always "unassociated alpha"
-    spec.attribute ("oiio:UnassociatedAlpha", (int)1);
+    // PNG files are always "unassociated alpha" but we convert to associated
+    // unless requested otherwise
+    if (keep_unassociated_alpha)
+        spec.attribute ("oiio:UnassociatedAlpha", (int)1);
 
     // FIXME -- look for an XMP packet in an iTXt chunk.
 }
@@ -304,14 +347,28 @@ create_write_struct (png_structp& sp, png_infop& ip, int& color_type,
         return "PNG does not support volume images (depth > 1)";
 
     switch (spec.nchannels) {
-    case 1 : color_type = PNG_COLOR_TYPE_GRAY; break;
-    case 2 : color_type = PNG_COLOR_TYPE_GRAY_ALPHA; break;
-    case 3 : color_type = PNG_COLOR_TYPE_RGB; break;
-    case 4 : color_type = PNG_COLOR_TYPE_RGB_ALPHA; break;
+    case 1 :
+        color_type = PNG_COLOR_TYPE_GRAY;
+        spec.alpha_channel = -1;
+        break;
+    case 2 :
+        color_type = PNG_COLOR_TYPE_GRAY_ALPHA;
+        spec.alpha_channel = 1;
+        break;
+    case 3 :
+        color_type = PNG_COLOR_TYPE_RGB;
+        spec.alpha_channel = -1;
+        break;
+    case 4 :
+        color_type = PNG_COLOR_TYPE_RGB_ALPHA;
+        spec.alpha_channel = 3;
+        break;
     default:
         return Strutil::format  ("PNG only supports 1-4 channels, not %d",
                                  spec.nchannels);
     }
+    // N.B. PNG is very rigid about the meaning of the channels, so enforce
+    // which channel is alpha, that's the only way PNG can do it.
 
     sp = png_create_write_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
     if (! sp)
@@ -361,7 +418,7 @@ put_parameter (png_structp& sp, png_infop& ip, const std::string &_name,
     if (Strutil::iequals(name, "DateTime") && type == TypeDesc::STRING) {
         png_time mod_time;
         int year, month, day, hour, minute, second;
-        if (sscanf (*(const char **)data, "%4d:%02d:%02d %2d:%02d:%02d",
+        if (sscanf (*(const char **)data, "%4d:%02d:%02d %02d:%02d:%02d",
                     &year, &month, &day, &hour, &minute, &second) == 6) {
             mod_time.year = year;
             mod_time.month = month;
@@ -419,7 +476,8 @@ put_parameter (png_structp& sp, png_infop& ip, const std::string &_name,
 ///
 inline void
 write_info (png_structp& sp, png_infop& ip, int& color_type,
-            ImageSpec& spec, std::vector<png_text>& text)
+            ImageSpec& spec, std::vector<png_text>& text,
+            bool& convert_alpha, float& gamma)
 {
     // Force either 16 or 8 bit integers
     if (spec.format == TypeDesc::UINT8 || spec.format == TypeDesc::INT8)
@@ -433,36 +491,57 @@ write_info (png_structp& sp, png_infop& ip, int& color_type,
 
     png_set_oFFs (sp, ip, spec.x, spec.y, PNG_OFFSET_PIXEL);
 
+    // PNG specifically dictates unassociated (un-"premultiplied") alpha
+    convert_alpha = spec.alpha_channel != -1 &&
+                    !spec.get_int_attribute("oiio:UnassociatedAlpha", 0);
+
+    gamma = spec.get_float_attribute ("oiio:Gamma", 1.0);
+
     std::string colorspace = spec.get_string_attribute ("oiio:ColorSpace");
     if (Strutil::iequals (colorspace, "Linear")) {
         png_set_gAMA (sp, ip, 1.0);
     }
-    else if (Strutil::iequals (colorspace, "GammaCorrected")) {
-        float gamma = spec.get_float_attribute ("oiio:Gamma", 1.0);
-        png_set_gAMA (sp, ip, gamma);
+    else if (Strutil::istarts_with (colorspace, "GammaCorrected")) {
+        float g = Strutil::from_string<float>(colorspace.c_str()+14);
+        if (g >= 0.01f && g <= 10.0f /* sanity check */)
+            gamma = g;
+        png_set_gAMA (sp, ip, 1.0f/gamma);
     }
     else if (Strutil::iequals (colorspace, "sRGB")) {
         png_set_sRGB_gAMA_and_cHRM (sp, ip, PNG_sRGB_INTENT_ABSOLUTE);
     }
-    
+
+    // Write ICC profile, if we have anything
+    const ParamValue* icc_profile_parameter = spec.find_attribute(ICC_PROFILE_ATTR);
+    if (icc_profile_parameter != NULL) {
+        unsigned int length = icc_profile_parameter->type().size();
+#if OIIO_LIBPNG_VERSION > 10500 /* PNG function signatures changed */
+        unsigned char *icc_profile = (unsigned char*)icc_profile_parameter->data();
+        if (icc_profile && length)
+            png_set_iCCP (sp, ip, "Embedded Profile", 0, icc_profile, length);
+#else
+        char *icc_profile = (char*)icc_profile_parameter->data();
+        if (icc_profile && length)
+            png_set_iCCP (sp, ip, (png_charp)"Embedded Profile", 0, icc_profile, length);
+#endif
+    }
+
     if (false && ! spec.find_attribute("DateTime")) {
         time_t now;
         time (&now);
         struct tm mytm;
         Sysutil::get_local_time (&now, &mytm);
-        std::string date = Strutil::format ("%4d:%02d:%02d %2d:%02d:%02d",
+        std::string date = Strutil::format ("%4d:%02d:%02d %02d:%02d:%02d",
                                mytm.tm_year+1900, mytm.tm_mon+1, mytm.tm_mday,
                                mytm.tm_hour, mytm.tm_min, mytm.tm_sec);
         spec.attribute ("DateTime", date);
     }
 
-    ImageIOParameter *unit=NULL, *xres=NULL, *yres=NULL;
-    if ((unit = spec.find_attribute("ResolutionUnit", TypeDesc::STRING)) &&
-        (xres = spec.find_attribute("XResolution", TypeDesc::FLOAT)) &&
-        (yres = spec.find_attribute("YResolution", TypeDesc::FLOAT))) {
-        const char *unitname = *(const char **)unit->data();
-        const float x = *(const float *)xres->data();
-        const float y = *(const float *)yres->data();
+    string_view unitname = spec.get_string_attribute ("ResolutionUnit");
+    float xres = spec.get_float_attribute ("XResolution");
+    float yres = spec.get_float_attribute ("YResolution");
+    float paspect = spec.get_float_attribute ("PixelAspectRatio");
+    if (xres || yres || paspect || unitname.size()) {
         int unittype = PNG_RESOLUTION_UNKNOWN;
         float scale = 1;
         if (Strutil::iequals (unitname, "meter") || Strutil::iequals (unitname, "m"))
@@ -474,8 +553,23 @@ write_info (png_structp& sp, png_infop& ip, int& color_type,
             unittype = PNG_RESOLUTION_METER;
             scale = 100.0/2.54;
         }
-        png_set_pHYs (sp, ip, (png_uint_32)(x*scale),
-                      (png_uint_32)(y*scale), unittype);
+        if (paspect) {
+            // If pixel aspect is given, allow resolution to be reset
+            if (xres)
+                yres = 0.0f;
+            else
+                xres = 0.0f;
+        }
+        if (xres == 0.0f && yres == 0.0f) {
+            xres = 100.0f;
+            yres = xres * (paspect ? paspect : 1.0f);
+        } else if (xres == 0.0f) {
+            xres = yres / (paspect ? paspect : 1.0f);
+        } else if (yres == 0.0f) {
+            yres = xres * (paspect ? paspect : 1.0f);
+        }
+        png_set_pHYs (sp, ip, (png_uint_32)(xres*scale),
+                      (png_uint_32)(yres*scale), unittype);
     }
 
     // Deal with all other params
@@ -540,7 +634,7 @@ destroy_write_struct (png_structp& sp, png_infop& ip)
 
 
 
-};  // namespace PNG_pvt
+}  // namespace PNG_pvt
 
 OIIO_PLUGIN_NAMESPACE_END
 
